/* vim: set noet:
 *
 * OpenTyrian Classic: A modern cross-platform port of Tyrian
 * Copyright (C) 2007  The OpenTyrian Development Team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "opentyr.h"
#include "loudness.h"
#include "varz.h"

#include "fm_synth.h"
#include "lds_play.h"
#include "params.h"


//Dreamcast code, to load WAV music!
SDL_AudioSpec wave;
SDL_AudioCVT cvt;
SDL_AudioCVT cvtsong31;
Uint8 *data;
Uint32 dlen;
char tmpstr[50];
#include "nortsong.h"
Uint8 *data2;
Uint32 dpos2;
Uint32 dlen2;
int oldsong;
bool end31song;
bool exclusio;

/* SYN: These are externally accessible variables: */
JE_MusicType musicData;
JE_boolean repeated;
extern JE_boolean playing;

float sample_volume = 0.4f;
float music_volume = 0.6f;

SDL_mutex *soundmutex = NULL;

/* SYN: These shouldn't be used outside this file. Hands off! */
SAMPLE_TYPE *channel_buffer[SFX_CHANNELS] = { NULL };
SAMPLE_TYPE *channel_pos[SFX_CHANNELS] = { NULL };
Uint32 channel_len[SFX_CHANNELS] = { 0 };
Uint8 channel_vol[SFX_CHANNELS];
int sound_init_state = false;
int freq = 11025 * OUTPUT_QUALITY;

bool music_playing = false;


void audio_cb(void *userdata, unsigned char *feedme, int howmuch);

/* SYN: The arguments to this function are probably meaningless now */
void JE_initialize(JE_word soundblaster, JE_word midi, JE_boolean mixenable, JE_byte sberror, JE_byte midierror)
{
	SDL_AudioSpec plz, got;
	int i = 0;

	if (SDL_InitSubSystem(SDL_INIT_AUDIO))
	{
		printf("Failed to initialize audio: %s\n", SDL_GetError());
		noSound = true;
		return;
	}

	sound_init_state = true;

	soundmutex = SDL_CreateMutex();

	if (soundmutex == NULL)
	{
		printf("Couldn't create mutex! Oh noes!\n");
		exit(-1);
	}

	opl_init();

	plz.freq = freq;
	plz.format = AUDIO_S16SYS;
	plz.channels = 1;
	plz.samples = 1024;
	plz.callback = audio_cb;
	plz.userdata = soundmutex;

	printf("\tRequested SDL frequency: %d; SDL buffer size: %d\n", plz.freq, plz.samples);

	if ( SDL_OpenAudio(&plz, NULL) < 0 )
	{
		printf("\tWARNING: Failed to initialize SDL audio. Bailing out.\n");
		exit(1);
	}

	printf("\tObtained  SDL frequency: %d; SDL buffer size: %d\n", plz.freq, plz.samples);
	
	end31song=false;
	exclusio=true;
	
	SDL_PauseAudio(0);
}

void audio_cb(void *userdata, unsigned char *sdl_buffer, int howmuch)
{
	static long ct = 0;

	SDL_mutex *mut = (SDL_mutex *) userdata;

	/* Making sure that we don't mess with sound buffers when someone else is using them! */
	if (SDL_mutexP(mut) == -1)
	{
		printf("Couldn't lock mutex! Argh!\n");
		exit(-1);
	}

	SAMPLE_TYPE *feedme = (SAMPLE_TYPE *) sdl_buffer;

	if (music_playing)
	{
		Uint32 size = (dlen2-dpos2);
		if ( size > (Uint32)howmuch ) size = howmuch;
		
		//loop
		if (dpos2-dlen2==0){ //end song... it doesn't matter if we try to set before it doesn't work....
			if (currentSong==10){
				//we don't want loop for these songs
			}
			else if (currentSong==31){
				end31song=true;  //no loop, open the function
			}
			else{ //loop song
				dlen2 = cvt.len_cvt;
				dpos2 = 0;
			}
		}
		
		if (currentSong==31){ //different buffer
			SDL_MixAudio( sdl_buffer, &cvtsong31.buf[dpos2], size, (int) (100*music_volume) );
		}
		else{
			SDL_MixAudio( sdl_buffer, &cvt.buf[dpos2], size, (int) (100*music_volume) );
		}
		dpos2 += size;
		}
		/* Reduce the music volume. */
		int qu = howmuch / BYTES_PER_SAMPLE;
		int smp = 0;
		for (smp; smp < qu; smp++)
		{
			feedme[smp] *= music_volume;
		}

	/* SYN: Mix sound channels and shove into audio buffer */
	int ch = 0;
	for (ch; ch < SFX_CHANNELS; ch++)
	{
		float volume = sample_volume * (channel_vol[ch] / 8.0f);
		
		/* SYN: Don't copy more data than is in the channel! */
		int qu = (howmuch > channel_len[ch] ? channel_len[ch] : howmuch) / BYTES_PER_SAMPLE;
		int smp = 0;
		for (smp; smp < qu; smp++)
		{
			long clip = (long)feedme[smp] + (long)(channel_pos[ch][smp] * volume);
			feedme[smp] = (clip > 0x7fff) ? 0x7fff : (clip <= -0x8000) ? -0x8000 : (short)clip;
		}

		channel_pos[ch] += qu;
		channel_len[ch] -= qu * BYTES_PER_SAMPLE;

		/* SYN: If we've emptied a channel buffer, let's free the memory and clear the channel. */
		if (channel_len[ch] == 0)
		{
			if(channel_buffer[ch]) free(channel_buffer[ch]);
			channel_buffer[ch] = channel_pos[ch] = NULL;
		}
	}

	SDL_mutexV(mut); /* release mutex */
}

void JE_deinitialize( void )
{
	/* SYN: Clean up any other audio stuff, if necessary. This should only be called when we're quitting. */
	int i = 0;
	for (i; i < SFX_CHANNELS; i++)
	{
		if(channel_buffer[i]) free(channel_buffer[i]);
		channel_buffer[i] = channel_pos[i] = NULL;
		channel_len[i] = 0;
	}
	
	opl_deinit();
	
	SDL_CloseAudio();
	SDL_QuitSubSystem(SDL_INIT_AUDIO);
}

void JE_play( void )
{
	/* SYN: This proc isn't necessary, because filling the buffer is handled in the SDL callback function.*/
}

/* SYN: selectSong is called with 0 to disable the current song. Calling it with 1 will start the current song if not playing,
   or restart it if it is. */
void JE_selectSong( JE_word value )
{
	if (noSound)
		return;

	SDL_LockAudio();
	/* TODO: The mutex'd region could possibly be smaller, but I wanted to keep it in this file. */

	/* Making sure that we don't mess with sound buffers when someone else is using them! */
	if (SDL_mutexP(soundmutex) == -1)
	{
		printf("Couldn't lock mutex! Argh!\n");
		exit(-1);
	}

	switch (value)
	{
		case 0: //first always calls stop (0), then is load/stop sequence...
			//We are going to free buffer whem we load the song!!!
			SDL_PauseAudio(1);
			music_playing = false;			
			break;
		case 1:
		case 2:
			if (currentSong==31){ //secret song, very small and can be loaded together
				//lds_load((JE_byte *) musicData); /* Load song */
				
				// Load Song from WAV File
				snprintf(tmpstr, 50, "/cd/data/song_%d.wav", currentSong);
				if ( SDL_LoadWAV(tmpstr, &wave, &data, &dlen) == NULL ) {
					printf("Error in SDL_LoadWAV.\n");
				}
				SDL_BuildAudioCVT( &cvtsong31, wave.format, wave.channels, wave.freq, AUDIO_S16SYS, 1, 22050 );
				cvtsong31.buf = (Uint8*)malloc( dlen*cvtsong31.len_mult );
				memcpy( cvtsong31.buf, data, dlen );
			
				cvtsong31.len = dlen;
				SDL_ConvertAudio(&cvtsong31);
				SDL_FreeWAV(data);
			
				data2 = cvtsong31.buf;
				dlen2 = cvtsong31.len_cvt;
				dpos2 = 0;
			}
			else{
				if (oldsong==31){
					//Use the song in memory!
					//First, free the buffer of the old song!
					if (cvtsong31.buf) free(cvtsong31.buf);
					cvtsong31.buf=NULL;
					data2 = cvt.buf;
					dlen2 = cvt.len_cvt;
					dpos2 = 0;
				}
				else{
				//buffer for song31 will be freeded in callback....
				//First, free the buffer of the song!
				if (cvt.buf) free(cvt.buf);
				cvt.buf=NULL;
					
				// Load Song from WAV File
				snprintf(tmpstr, 50, "/cd/data/song_%d.wav", currentSong);
				if ( SDL_LoadWAV(tmpstr, &wave, &data, &dlen) == NULL ) {
					printf("Error in SDL_LoadWAV.\n");
				}

				SDL_BuildAudioCVT( &cvt, wave.format, wave.channels, wave.freq, AUDIO_S16SYS, 1, 22050 );
				cvt.buf = (Uint8*)malloc( dlen*cvt.len_mult );
				memcpy( cvt.buf, data, dlen );
			
				cvt.len = dlen;
				SDL_ConvertAudio(&cvt);
				SDL_FreeWAV(data);
		
				data2 = cvt.buf;
				dlen2 = cvt.len_cvt;
				dpos2 = 0;
				}
			}
			oldsong=currentSong;
			SDL_PauseAudio(0);
			music_playing = true;
			break;
		default:
			printf("JE_selectSong: fading TODO!\n");
			/* TODO: Finish this FADING function! */
			break;
	}

	SDL_mutexV(soundmutex); /* release mutex */
	SDL_UnlockAudio();
}

void JE_samplePlay(JE_word addlo, JE_word addhi, JE_word size, JE_word freq)
{
	/* SYN: I don't think this function is used. */
	STUB();
}

void JE_bigSamplePlay(JE_word addlo, JE_word addhi, JE_word size, JE_word freq)
{
	/* SYN: I don't think this function is used. */
	STUB();
}

/* Call with 0x1-0x100 for music volume, and 0x10 to 0xf0 for sample volume. */
/* SYN: Either I'm misunderstanding Andreas's comments, or the information in them is inaccurate. */
void JE_setVol(JE_word volume, JE_word sample)
{
	/* printf("JE_setVol: music: %d, sample: %d\n", volume, sample); */
	
	if (volume > 0)
		music_volume = volume * (float)(0.6 / 256.0);
	sample_volume = sample * (float)(0.4 / 256.0);
}

JE_word JE_getVol( void )
{
	STUB();
	return 0;
}

JE_word JE_getSampleVol( void )
{
	STUB();
	return 0;
}

void JE_multiSampleInit(JE_word addlo, JE_word addhi, JE_word dmalo, JE_word dmahi)
{
	/* SYN: I don't know if this function should do anything else. For now, it just checks to see if sound has
	   been initialized and, if not, calls the main initialize function. */

	if (!sound_init_state)
	{
		JE_initialize(0, 0, 0, 0, 0);
	}
}

void JE_multiSampleMix( void )
{
	/* SYN: This proc isn't necessary, because the mixing is handled in the SDL callback function.*/
}

void JE_multiSamplePlay(JE_byte *buffer, JE_word size, JE_byte chan, JE_byte vol)
{
	if (noSound)
		return;
	
	SDL_LockAudio();
	
	/* Making sure that we don't mess with sound buffers when someone else is using them! */
	if (SDL_mutexP(soundmutex) == -1)
	{
		printf("Couldn't lock mutex! Argh!\n");
		exit(-1);
	}

	if (channel_buffer[chan]) free(channel_buffer[chan]);

	channel_len[chan] = size * BYTES_PER_SAMPLE * SAMPLE_SCALING;
	channel_buffer[chan] = malloc(channel_len[chan]);
	channel_pos[chan] = channel_buffer[chan];
	channel_vol[chan] = vol + 1;
	int i = 0;
	for (i; i < size; i++)
	{
		int ex = 0;
		for (ex; ex < SAMPLE_SCALING; ex++)
		{
			channel_buffer[chan][(i * SAMPLE_SCALING) + ex] = ((SAMPLE_TYPE) ((Sint8) buffer[i]) << 8);
		}
	}

	SDL_mutexV(soundmutex); /* release mutex */
	SDL_UnlockAudio();
}

void checkend31song(void)
{
	if (end31song){
		if (exclusio){
			exclusio=false;
			end31song=false;
			JE_playSong(levelSong);
			exclusio=true;
		}
	}
}

